program ANacousticG2d
  ! Program to calculate Green's function of acoustic wave equation in 2d
  ! Function of radius only
  ! See SEP138 page 56

  use sep
  implicit none
  
  integer              :: nr,ir
  real                 :: or,dr,r
  
  complex, allocatable :: dataw(:)
  real                 :: v
  
  real                 :: f,w
  real,parameter       :: pi = 3.14159265
  
  ! Header IO
  call sep_init()
  
  ! Geometry
  call from_param('dr',dr,  1.  )
  call from_param('or',or,  0.  )
  call from_param('nr',nr,  100 )

  ! Read velocity
  call from_param('v' , v, 1500.)

  ! Read frequency information
  call from_param('f',f,     5.)
  call from_param('w',w,2.*pi*f)

  ! Read velocity
  call to_history("n1",nr)
  call to_history("d1",dr)
  call to_history("o1",or)
  call to_history("label1",'Receiver')
  call to_history("esize",8)
  
  call sep_close()
  
  ! Allocate data-space and source-vector
  allocate( dataw(nr) ); dataw  = 0.
  
  ! Add common part of Green's function
  dataw = 1./sqrt(8.*pi*w/v)*cexp(cmplx(0.,-pi/4.))
  
  ! Convert to source type (and add smoothing weight)
  do ir=1,nr
    
    ! Distance
    r   = abs(or+(ir-1)*dr)
    
    ! Solution
    if (r.ne.0.) dataw(ir) = 1./sqrt(r)*cexp(cmplx(0.,-w/v*r))*dataw(ir)
    
  end do
  
  ! Write data
  call srite("out",dataw,8*size(dataw));
  
  ! Deallocate and close
  deallocate( dataw );
  call exit(0);

end program


