program FForward

  use sep

  implicit none

  integer                             :: dev_num, ft_j
  integer                             :: nsx, fsx, jsx, nrx, frx, jrx, irx, isx
  integer                             :: nsz, fsz, jsz, nrz, frz, jrz, irz, isz
  integer                             :: snt,nst,nft,pnt,to,ct
  integer                             :: crop1, crop2, fat, nwl
  real                                :: dt, dz, oz, dx, ox
  integer                             :: it, ix, iz, ir, is
  integer                             :: nt, nx, nz, nr, ns
  
  real                                :: rho,s_fact
  real, dimension(:,:), allocatable   :: vel0
  real, dimension(:)  , allocatable   :: sinus
  real, dimension(:) , allocatable    :: s_amp,r_real,r_imag
  integer, dimension(:), allocatable  :: s_phs
  integer,dimension(:,:),allocatable  :: rec_zx,sou_zx,src_zx

  real, dimension(:), allocatable     :: src_amp
  integer, dimension(:), allocatable  :: src_phs

  real                                :: w,f
  real,parameter                      :: pi=3.1415926535

  integer, dimension(8)               :: timer0,timer1
  real                                :: timer_sum=0.

  ! Smaller computational domain
  integer                             :: n1,n2

  call sep_init()

  !---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
  !---------------------------------------------------------------------------
  
  ! Read background velocity geometry
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)
  
  if(mod(nz-10,16) .ne. 0) call erexit("(n1-10)/20 is not an integer")
  if(mod(nx-10,16) .ne. 0) call erexit("(n2-10)/20 is not an integer")
  
  ! Read frequency information
  call from_param('f',f,     5.)
  call from_param('w',w,2.*pi*f)
  call from_param('ft_j',ft_j, 1)
  
  ! Read in Density
  call from_param('rho', rho, 1000.0)
  call from_param("crop1",crop1,48)
  call from_param("crop2",crop2,48)
  call from_param("dev_num",dev_num,1)
  fat=5  ! Stencil fat
  
  ! Read shots geometry
  if (exist_file('sou_in')) then
    call from_aux('sou_in','n1',ns,1)
    allocate(src_zx( 1,2))
    allocate(sou_zx(ns,2))
    call sreed('sou_in',sou_zx,4*size(sou_zx))
  else
    call from_param("nsx", nsx,     1)
    call from_param("fsx", fsx, crop1)
    call from_param("jsx", jsx,     1)
    call from_param("nsz", nsz,     1)
    call from_param("fsz", fsz, crop2)
    call from_param("jsz", jsz,     1)
    ns = nsx*nsz
    ! Prepare source indices   (c-style)
    ns = nsx*nsz
    allocate(src_zx( 1,2))
    allocate(sou_zx(ns,2))
    do isx=1,nsx
    do isz=1,nsz
      sou_zx((isz-1)*nsz+isx,:) = (/ fsz+(isz-1)*jsz , fsx+(isx-1)*jsx /)
    end do
    end do
  end if

  ! Read receivers geometry
  if (exist_file('rec_in')) then
    call from_aux('rec_in','n1',nr,1)
    allocate(rec_zx(nr,2))
    call sreed('rec_in',rec_zx,4*size(rec_zx))
  else
    call from_param("nrx", nrx,     1)
    call from_param("frx", frx, crop1)
    call from_param("jrx", jrx,     1)
    call from_param("nrz", nrz,     1)
    call from_param("frz", frz, crop2)
    call from_param("jrz", jrz,     1)
    nr = nrx*nrz
    ! Prepare receiver indices (c-style)
    allocate(rec_zx(nr,2))
    do irx=1,nrx
    do irz=1,nrz
      rec_zx((irz-1)*nrz+irx,:) = (/ frz+(irz-1)*jrz , frx+(irx-1)*jrx /)
    end do
    end do
  end if

  ! Read in the background velocity
  allocate( vel0(nz,nx) )
  call sep_read(vel0)
  
  
  ! FD Time stepping scheme and source function
  call from_param('dt' ,dt , 0.)
  call from_param('nst',nst, 0 )
  if (dt.eq.0.) then        ! Fine tune time stepping issues (John Etgen steps)
    dt = .333*min(dx,dz)/maxval(vel0(:,:))
 !   write(0,*) 'dt before rounding',dt
    pnt= nint(1./f/dt/4)*4  ! In 1 wavelength, be dividable by 4
 !   write(0,*) 'pnt',pnt
    dt =      1./f/pnt      ! So thus
 !   write(0,*) 'dt after rounding',dt
  end if
  call from_param('nwl',nwl, 10 )
  nft= nwl*pnt              ! Simulate 10 wavelengths
  
  if (nst.eq.0) then        ! time it takes to travel diagonaly through model space
    nst = nint(sqrt(((nx-2*crop1)*dx)**2.+((nz-2*crop2)*dz)**2.)/minval(vel0(:,:))/dt/pnt+1)*pnt
  else
    nst = nint(real(nst)/pnt)*pnt
  end if
  nt = nst + nft
  ct = pnt/4
  snt = pnt+pnt+ct+1
  
  
  write(0,*) "Calculation parameters"
  write(0,*) 'dt,pnt,nst,nft,nt,snt'
  write(0,*)  dt,pnt,nst,nft,nt,snt
  
  ! Allocation
  allocate( s_phs(ns), s_amp(ns), r_imag(nr), r_real(nr), sinus(snt), src_phs(1), src_amp(1) )
  
  
  call to_history("o1", 1 )
  call to_history("d1", 1 )
  call to_history("n1", nr)
  call to_history("o2", 1 )
  call to_history("d2", 1 )
  call to_history("n2", ns)
  call to_history("o3", 1 )
  call to_history("d3", 1 )
  call to_history("n3", 1 )
  call to_history("label1", "Receivers")
  call to_history("label2", "Sources"  )
  call to_history("label3",    " "     )
  call to_history("esize" ,8)
  
  



  
  ! Write these sources and receivers to file for later use
  if (exist_file('sou_hdr')) then
    call to_history("n1",ns,"sou_hdr")
    call to_history("n2",2 ,"sou_hdr")
    call to_history("d1",1.,"sou_hdr")
    call to_history("d2",1.,"sou_hdr")
    call to_history("o1",1.,"sou_hdr")
    call to_history("o2",1.,"sou_hdr")
    call to_history("label1","source number","sou_hdr")
    call to_history("label2","(z,x) index"  ,"sou_hdr")
    call srite('sou_hdr',sou_zx,4*2*ns)
  end if
  
  ! Write these sources and receivers to file for later use
  if (exist_file('rec_hdr')) then
    call to_history("n1",nr,"rec_hdr")
    call to_history("n2",2 ,"rec_hdr")
    call to_history("d1",1.,"rec_hdr")
    call to_history("d2",1.,"rec_hdr")
    call to_history("o1",1.,"rec_hdr")
    call to_history("o2",1.,"rec_hdr")
    call to_history("label1","receiver number","rec_hdr")
    call to_history("label2","(z,x) index"    ,"rec_hdr")
    call srite('rec_hdr',rec_zx,4*2*nr)
  end if
  
  !write(0,*) "Fortran here"
  !write(0,*) "z receiver indices",rec_zx(:,1)
  !write(0,*) "x receiver indices",rec_zx(:,2)
  !write(0,*) "z source indices",  sou_zx(:,1)
  !write(0,*) "x source indices",  sou_zx(:,2)
  !write(0,*) "================="

  ! Create base sinus
  do it=1,snt
    sinus(it)=sin(w*((it-1-pnt/2)*dt))
  end do
  
  ! Prepare source amplitude and phase
  s_fact = dt*dt/(dx*dz)/pi
  do is=1,ns
    s_phs(is) = 0
    s_amp(is) = s_fact * vel0(sou_zx(is,1),sou_zx(is,2))**2
  end do
  
  ! Init GPU modules
  call c_init(pnt, nst, nft, dt, nz, nx, dz, dx, dev_num, sinus)
  call c_init_model(nr, 1)  ! Model one source at a time
  call c_vel_update(vel0*dt/dx/4., vel0*vel0)
  

  ! Loop over sources
  do is=1,ns
    write(0,*) "Shot",is,'out of',ns
    
    ! Pass single source location
    src_zx (1,:) = sou_zx(is,:)
    src_amp(1  ) = s_amp (is  )
    src_phs(1  ) = s_phs (is  )


    ! Model data
    write(0,*) 'Call modelling'
    call c_model(src_zx, rec_zx, src_amp(1  ), src_phs(1  ), r_real, r_imag, ft_j)
    r_real = ft_j*dt*r_real
    r_imag = ft_j*dt*r_imag
    write(0,*) 'End modelling'
    
    ! Write data to disk
    call srite("out",cmplx(r_real, r_imag),8*nr)
    
  end do
  
  !---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  !---------------------------------------------------------------------------
  !---------------------------------------------------------------------------
  ! call to_history("timer sum (sec)", timer_sum/1000)
  ! call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
  !---------------------------------------------------------------------------
  call sep_close()
  call exit(0)
 
end program
