program FWI

  use sep
  use tomo_mod

  implicit none

  integer                              :: dev_num, ft_mod_j, ft_grd_j
  integer                              :: snt,nst,nft,pnt,to,ct
  integer                              :: n1s,n2s,crop1,crop2,stat,fat,nwl,ismall,n3
  real                                 :: dt, dz, oz, dx, ox
  integer                              :: it, ix, iz, ir, is
  integer                              :: nt, nx, nz, nr, ns

  integer                              :: iter, niter, itry, ntry, counter
  real                                 :: parab_a, parab_b

  real                                 :: rho, s_fact
  real, dimension(:), allocatable      :: sinus
  real, dimension(:,:), allocatable    :: slowsq0, slowsq, vel0, vel, gradient0, gradient, search

  real, dimension(:) , allocatable     :: r_real,r_imag
  real, dimension(:) , allocatable     :: s_amp
  integer, dimension(:), allocatable   :: s_phs

  real, dimension(:), allocatable      :: src_amp
  integer, dimension(:), allocatable   :: src_phs
  
  real, dimension(:,:), allocatable    :: s_wfld_real, s_wfld_imag
  real, dimension(:,:), allocatable    :: r_wfld_real, r_wfld_imag
  real, dimension(:,:), allocatable    :: d_obs_real , d_obs_imag
  real, dimension(:), allocatable      :: d_diff_real, d_diff_imag
  real, dimension(:), allocatable      :: d_diff_amp
  integer, dimension(:), allocatable   :: d_diff_phs

  complex, dimension(:,:), allocatable :: d_obs

  double precision                     :: r0, r1, r2, r3, step_size, beta
  real                                 :: slowsq_min, update_max, update_scale
  integer                              :: sou_x, rec_x, irec

  integer, dimension(8)                :: timer0, timer1
  real                                 :: timer_sum=0.

  integer,dimension(:,:),allocatable   :: rec_zx,sou_zx,src_zx

  real                                 :: w,f
  real,parameter                       :: pi=3.1415926535

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)

  ! Read shots geometry
  if (exist_file('sou_hdr')) then
    call from_aux("sou_hdr","n1",ns,1)
    allocate(sou_zx(ns,2))
    call sreed('sou_hdr',sou_zx,4*size(sou_zx))
  else
    call erexit("Missing shot geometry file")
  end if
  allocate(src_zx( 1,2))

  ! Read receiver geometry
  if (exist_file('rec_hdr')) then
    call from_aux("rec_hdr","n1",nr,1)
    allocate(rec_zx(nr,2))
    call sreed('rec_hdr',rec_zx,4*size(rec_zx))
  else
    call erexit("Missing receiver geometry file")
  end if

  ! Read frequency information
  call from_param('f',f,     5.)
  call from_param('w',w,2.*pi*f)
  call from_param('ft_mod_j',ft_mod_j, 1)
  call from_param('ft_grd_j',ft_grd_j, 1)
  


  ! Read in Density
  call from_param('rho', rho, 1000.0)
  call from_param("crop1",crop1,48)
  call from_param("crop2",crop2,48)
  call from_param("dev_num",dev_num,1)
  fat=5  ! Stencil fat

  call from_param('niter', niter, 1)
  call from_param('ntry' , ntry , 1)

  n1s = nz-2*crop1
  n2s = nx-2*crop2


  ! Read in the background velocity
  allocate( vel0(nz,nx), vel(nz,nx), slowsq0(nz,nx), slowsq(nz,nx) )
  call sep_read(vel0)
  slowsq0 = 1./(vel0**2)
  vel = 0.; slowsq = 0.

  ! FD Time stepping scheme and source function
  call from_param('dt' ,dt , 0.)
  call from_param('nst',nst, 0 )
  if (dt.eq.0.) then        ! Fine tune time stepping issues (John Etgen steps)
    dt = .333*min(dx,dz)/maxval(vel0(:,:))
    write(0,*) 'dt before rounding',dt
    pnt= nint(1./f/dt/4)*4  ! In 1 wavelength, be dividable by 4
    write(0,*) 'pnt',pnt
    dt =      1./f/pnt      ! So thus
    write(0,*) 'dt after rounding',dt
  end if
  call from_param('nwl',nwl, 10 )
  nft= nwl*pnt              ! Simulate 10 wavelengths
  
  if (nst.eq.0) then        ! time it takes to travel diagonaly through model space
    nst = nint(sqrt(((nx-2*crop1)*dx)**2.+((nz-2*crop2)*dz)**2.)/minval(vel0(:,:))/dt/pnt+1)*pnt
  else
    nst = nint(real(nst)/pnt)*pnt
  end if
  nt = nst + nft
  ct = pnt/4
  snt = pnt+pnt+ct+1
  
  
  write(0,*) "Calculation parameters"
  write(0,*) 'dt,pnt,nst,nft,nt,snt'
  write(0,*)  dt,pnt,nst,nft,nt,snt
  write(0,*) "nr,ns",nr,ns

  ! Read in the observed data
  allocate( d_obs(nr,ns) )
  allocate( d_obs_real(nr,ns), d_obs_imag(nr,ns) )
  call sreed("data_obs",d_obs,8*size(d_obs))  
  d_obs_real =  real( d_obs )
  d_obs_imag = aimag( d_obs )
  deallocate( d_obs )

  ! Create base sinus
  allocate( sinus(snt) )
  do it=1,snt
    sinus(it)=sin(w*((it-1-pnt/2)*dt))
  end do
  
  ! Prepare source amplitude and phase
  s_fact = dt*dt/(dx*dz)/pi
  allocate(s_phs(ns),s_amp(ns))
  do is=1,ns
    s_phs(is) = 0
    s_amp(is) = s_fact * vel0(sou_zx(is,1),sou_zx(is,2))**2
  end do
  call to_history("n1",  nz)
  call to_history("o1",  oz)
  call to_history("d1",  dz)
  call to_history("n2",  nx)
  call to_history("o2",  ox)
  call to_history("d2",  dx)
  call to_history("o3", 0.0)
  call to_history("d3", 1.0)


 ! write(0,*) "Fortran here"
 ! write(0,*) "z receiver indices",rec_zx(:,1)
 ! write(0,*) "x receiver indices",rec_zx(:,2)
 ! write(0,*) "z source indices",  sou_zx(:,1)
 ! write(0,*) "x source indices",  sou_zx(:,2)
 ! write(0,*) "================="


  ! More Allocation
  allocate( gradient0(nz,nx), gradient(nz,nx), search(nz,nx) )
  allocate( s_wfld_real(n1s,n2s), r_wfld_real(n1s,n2s) )
  allocate( s_wfld_imag(n1s,n2s), r_wfld_imag(n1s,n2s) )
  allocate( d_diff_real(nr)     , d_diff_imag(nr)      )
  allocate( d_diff_phs (nr)     , d_diff_amp (nr)      )
  allocate(    src_phs (1 )     ,    src_amp (1 )      )
  

  ! Init GPU modules
  call c_init(pnt, nst, nft, dt, nz, nx, dz, dx, dev_num, sinus)
  call c_init_prop(nr, 1, n1s, n2s, crop1, crop2 )  ! Model one source at a time

  ! Optimization iterations
  counter = 0; n3=0
  do iter=1,niter
    write(0,*) 'iteration',iter



    ! Iterate gradient
    if(iter > 1) gradient0 = gradient
    
    ! upload velocity to GPU
    call c_vel_update(vel0*dt/dx/4., vel0*vel0)
    
    write(0,*) "Calculate gradient"
    r0=0
    gradient = 0.
    do is=1,ns
      ! Single source
      src_zx (1,:) = sou_zx(is,:)
      src_amp(1  ) = s_amp (is  )
      src_phs(1  ) = s_phs (is  )
      
      d_diff_real = 0.; s_wfld_real = 0.; r_wfld_real = 0.
      d_diff_imag = 0.; s_wfld_imag = 0.; r_wfld_imag = 0.
      
      ! Forward computation
      call c_prop_for(src_zx,    src_amp,    src_phs, s_wfld_real , s_wfld_imag , ft_grd_j)
      s_wfld_real = real(ft_grd_j)*dt*s_wfld_real
      s_wfld_imag = real(ft_grd_j)*dt*s_wfld_imag
      
      ! Difference of source wavefield with observed data at receivers
      do ir=1,nr
        d_diff_real(ir) = s_wfld_real(rec_zx(ir,1)-crop1+1,rec_zx(ir,2)-crop2+1) - d_obs_real(ir,is)
        d_diff_imag(ir) = s_wfld_imag(rec_zx(ir,1)-crop1+1,rec_zx(ir,2)-crop2+1) - d_obs_imag(ir,is)
      end do
      
      ! Residual
      r0 = r0 + sum(dprod(d_diff_real,d_diff_real))+ sum(dprod(d_diff_imag,d_diff_imag))
      
      ! Data difference amplitude and phase
      d_diff_phs  =  nint(atan2(d_diff_imag,         d_diff_real)/pi*pnt/2)
      d_diff_amp  =       abs  (d_diff_imag) + abs  (d_diff_real)
      
      ! compute source wavefield
      s_wfld_real = -w**2*s_wfld_real
      s_wfld_imag = -w**2*s_wfld_imag
      
      ! compute receiver wavefield
      call c_prop_adj(rec_zx, d_diff_amp, d_diff_phs, r_wfld_real , r_wfld_imag , ft_grd_j)
      r_wfld_real = real(ft_grd_j)*dt*r_wfld_real
      r_wfld_imag = real(ft_grd_j)*dt*r_wfld_imag
      
      ! compute grdient
      gradient(1+crop1:nz-crop1,1+crop2:nx-crop2) = &
      gradient(1+crop1:nz-crop1,1+crop2:nx-crop2) - s_wfld_real*r_wfld_real + s_wfld_imag*r_wfld_imag
    end do ! end of shot loop
    
    ! Extend gradient
    call extend(gradient, nz, nx, crop1, crop2)
    
    ! Exit if residual is zero
    if(r0 .eq. 0.0) then
      write(0,*) "r0 is zero, will exit the program"
      exit
    end if

    ! Search direction
    search = -gradient
    
    ! First guess of step size, keep it small enough
    slowsq_min = minval(slowsq0)
    update_max = maxval(abs(search))
    update_scale = slowsq_min / update_max / 50.
    if (iter .eq. 1) step_size = slowsq_min / update_max / 100.
    
    ! Search
    do itry=1,ntry
      write(0,*) "Find step size, attempt",itry
      
      ! Reset step size if optimization keeps failing
      if(itry .eq. ntry/2) then
         step_size = slowsq_min / update_max / 100.
      endif

      write(0,*) "Test step 1"
      slowsq = slowsq0 + search * step_size * 0.5
      vel = 1./sqrt(slowsq)

      ! upload velocity and absorbing coeffs velocity to GPU
      call c_vel_update(vel*dt/dx/4., vel*vel)
      
      r1=0
      do is=1,ns
        ! Single source
        src_zx (1,:)=sou_zx(is,:); src_amp(1)=s_amp (is); src_phs(1)=s_phs(is)
        d_diff_real = 0.; d_diff_imag = 0.
        call c_model(src_zx, rec_zx,  s_amp, s_phs, d_diff_real, d_diff_imag, ft_mod_j)
        d_diff_real = real(ft_mod_j)*dt*d_diff_real - d_obs_real(:,is)
        d_diff_imag = real(ft_mod_j)*dt*d_diff_imag - d_obs_imag(:,is)
        ! Forward computation
        call c_prop_for(src_zx,    src_amp,    src_phs, s_wfld_real , s_wfld_imag , ft_grd_j)
        s_wfld_real = real(ft_grd_j)*dt*s_wfld_real
        s_wfld_imag = real(ft_grd_j)*dt*s_wfld_imag
        ! Residual
        r1 = r1 + sum(dprod(d_diff_real,d_diff_real))+ sum(dprod(d_diff_imag,d_diff_imag))
      end do
      
      
      write(0,*) "Test step 2"
      slowsq = slowsq0 + search * step_size * 1.0
      vel = 1./sqrt(slowsq)

      ! upload velocity and absorbing coeffs velocity to GPU
      call c_vel_update(vel*dt/dx/4., vel*vel)

      r2=0
      do is=1,ns
        ! Single source
        src_zx (1,:)=sou_zx(is,:); src_amp(1)=s_amp (is); src_phs(1)=s_phs(is)
        d_diff_real = 0.; d_diff_imag = 0.
        call c_model(src_zx, rec_zx,  s_amp, s_phs, d_diff_real, d_diff_imag, ft_mod_j)
        d_diff_real = real(ft_mod_j)*dt*d_diff_real - d_obs_real(:,is)
        d_diff_imag = real(ft_mod_j)*dt*d_diff_imag - d_obs_imag(:,is)
        ! Forward computation
        call c_prop_for(src_zx,    src_amp,    src_phs, s_wfld_real , s_wfld_imag , ft_grd_j)
        s_wfld_real = real(ft_grd_j)*dt*s_wfld_real
        s_wfld_imag = real(ft_grd_j)*dt*s_wfld_imag
        ! Residual
        r2 = r2 + sum(dprod(d_diff_real,d_diff_real))+ sum(dprod(d_diff_imag,d_diff_imag))
      end do
      
      
      write(0,*) "Test step 3"
      slowsq = slowsq0 + search * step_size * 1.5
      vel = 1./sqrt(slowsq)

      ! upload velocity and absorbing coeffs velocity to GPU
      call c_vel_update(vel*dt/dx/4., vel*vel)

      r3=0
      do is=1,ns
        ! Single source
        src_zx (1,:)=sou_zx(is,:); src_amp(1)=s_amp (is); src_phs(1)=s_phs(is)
        d_diff_real = 0.; d_diff_imag = 0.
        call c_model(src_zx, rec_zx,  s_amp, s_phs, d_diff_real, d_diff_imag, ft_mod_j)
        d_diff_real = real(ft_mod_j)*dt*d_diff_real - d_obs_real(:,is)
        d_diff_imag = real(ft_mod_j)*dt*d_diff_imag - d_obs_imag(:,is)
        ! Forward computation
        call c_prop_for(src_zx,    src_amp,    src_phs, s_wfld_real , s_wfld_imag , ft_grd_j)
        s_wfld_real = real(ft_grd_j)*dt*s_wfld_real
        s_wfld_imag = real(ft_grd_j)*dt*s_wfld_imag
        ! Residual
        r3 = r3 + sum(dprod(d_diff_real,d_diff_real))+ sum(dprod(d_diff_imag,d_diff_imag))
      end do
      
      write(0,*) "residual =", real(r0), real(r1), real(r2), real(r3)
      call stepper(r0, r1, r2, r3, step_size, ismall, stat)
      
      if(stat .eq. 0) exit
      
    enddo
    
    if(stat .ne. 0) then
      write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
      exit
    endif
    
    if(ismall .eq. 1) then
      r0 = r1
    elseif(ismall .eq. 2) then
      r0 = r2
    elseif(ismall .eq. 3) then
      r0 = r3
    else
      write(0,*) "something wrong with ismall, will exit the program"
      exit
    endif
    
    slowsq0= slowsq0 + step_size * search
    vel0 = 1./sqrt(slowsq0)
    call srite('out',vel0,4*size(vel0))
    n3=n3+1
    
  end do ! end of iteration loop


  ! For now write out gradient

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------
  call to_history("n3",n3)
  call sep_close()
  call exit(0)
end program
