program Geometry
  use sep
  implicit none
  
  
  integer                             :: nsx, fsx, jsx, nrx, frx, jrx, irx, isx
  integer                             :: nsz, fsz, jsz, nrz, frz, jrz, irz, isz
  
  integer,dimension(:,:),allocatable  :: rec_zx, sou_zx, rec_zx_old, sou_zx_old

  integer                             :: nr_old,ns_old,ns,nr,crop1,crop2,is,ir
  
  
  call sep_init()
  
  if (exist_file('sou_in')) then
    call from_aux('sou_in','n1',ns_old,1)
    allocate(sou_zx_old(ns_old,2))
    call sreed( 'sou_in',sou_zx_old,4*size(sou_zx_old))
  else
    ns_old=0
  end if

  if (exist_file('rec_in')) then
    call from_aux('rec_in','n1',nr_old,1)
    allocate(rec_zx_old(nr_old,2))
    call sreed( 'rec_in',rec_zx_old,4*size(rec_zx_old))
  else
    nr_old=0
  end if
  
  
  
  call from_param("crop1",crop1,0)
  call from_param("crop2",crop2,0)
  
 
  ! Read shots geometry
  call from_param("nsx", nsx,     1)
  call from_param("fsx", fsx, crop1)
  call from_param("jsx", jsx,     1)
  call from_param("nsz", nsz,     1)
  call from_param("fsz", fsz, crop2)
  call from_param("jsz", jsz,     1)
  ns = nsx*nsz
  ! Read receivers geometry
  call from_param("nrx", nrx,     1)
  call from_param("frx", frx, crop1)
  call from_param("jrx", jrx,     1)
  call from_param("nrz", nrz,     1)
  call from_param("frz", frz, crop2)
  call from_param("jrz", jrz,     1)
  nr = nrx*nrz

  write(0,*) nsx,fsx,jsx, nsz,fsz,jsz
  write(0,*) nrx,frx,jrx, nrz,frz,jrz
  
 
  ! Prepare receiver indices (c-style)
  allocate(rec_zx(nr,2))
  ir=0
  do irx=1,nrx
  do irz=1,nrz
    ir = ir+1
    rec_zx(ir,:) = (/ frz+(irz-1)*jrz , frx+(irx-1)*jrx /)
  end do
  end do
  
  ! Prepare source indices   (c-style)
  ns = nsx*nsz
  allocate(sou_zx(ns,2))
  is=0
  do isx=1,nsx
  do isz=1,nsz
    is = is+1
    sou_zx(is,:) = (/ fsz+(isz-1)*jsz , fsx+(isx-1)*jsx /)
  end do
  end do
  

  ! Write these sources and receivers to file for later use
  if (exist_file('sou_hdr')) then
    call to_history("n1",ns+ns_old,"sou_hdr")
    call to_history("n2",2 ,"sou_hdr")
    call to_history("d1",1.,"sou_hdr")
    call to_history("d2",1.,"sou_hdr")
    call to_history("o1",1.,"sou_hdr")
    call to_history("o2",1.,"sou_hdr")
    call to_history("label1","source number","sou_hdr")
    call to_history("label2","index"  ,"sou_hdr")
  end if

  ! Write these sources and receivers to file for later use
  if (exist_file('rec_hdr')) then
    call to_history("n1",nr+nr_old,"rec_hdr")
    call to_history("n2",2 ,"rec_hdr")
    call to_history("d1",1.,"rec_hdr")
    call to_history("d2",1.,"rec_hdr")
    call to_history("o1",1.,"rec_hdr")
    call to_history("o2",1.,"rec_hdr")
    call to_history("label1","receiver number","rec_hdr")
    call to_history("label2","index"    ,"rec_hdr")
  end if
  
  if (exist_file('rec_in')) call srite( 'rec_hdr',rec_zx_old(:,1),4*size(rec_zx_old(:,1)))
                            call srite( 'rec_hdr',rec_zx    (:,1),4*size(rec_zx    (:,1)))
  if (exist_file('rec_in')) call srite( 'rec_hdr',rec_zx_old(:,2),4*size(rec_zx_old(:,2)))
                            call srite( 'rec_hdr',rec_zx    (:,2),4*size(rec_zx    (:,2)))
  if (exist_file('sou_in')) call srite( 'sou_hdr',sou_zx_old(:,1),4*size(sou_zx_old(:,1)))
                            call srite( 'sou_hdr',sou_zx    (:,1),4*size(sou_zx    (:,1)))
  if (exist_file('sou_in')) call srite( 'sou_hdr',sou_zx_old(:,2),4*size(sou_zx_old(:,2)))
                            call srite( 'sou_hdr',sou_zx    (:,2),4*size(sou_zx    (:,2)))
  
  deallocate(sou_zx,rec_zx)
  if (exist_file('sou_in')) deallocate(sou_zx_old)
  if (exist_file('rec_in')) deallocate(rec_zx_old)
  call sep_close()
  call exit(0)

end program
