#include "cudaErrors.cu"

#define dloc(z,x) (z)+(x)*n1gpu

__constant__ int pnt_gpu;
__constant__ int nft_gpu;
__constant__ int nst_gpu;
__constant__ int nt_gpu;
__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int nrec_gpu;
__constant__ int nsrc_gpu;
__constant__ int crop1_gpu;
__constant__ int crop2_gpu;
__constant__ int n1s_gpu;
__constant__ int n2s_gpu;
texture <float, 2, cudaReadModeElementType>  vel_ref;
texture <float, 2, cudaReadModeElementType> vel2_ref;
cudaChannelFormatDesc cf;
cudaArray *vel_array=0;
cudaArray *vel2_array=0;

float *p0_gpu, *p1_gpu, *p2_gpu, *p3_gpu, *w_imag_gpu, *w_real_gpu, *sinus_gpu;
float *r_amp_gpu, *s_amp_gpu, *r_real_gpu, *r_imag_gpu;
int   *r_phs_gpu, *s_phs_gpu, *rec_zx_gpu, *src_zx_gpu;

#include "wave_kernel2d10o.cu"

__global__ void inject_kernel_src(int it, int ct, float *s_amp_gpu, int *s_phs_gpu, float *p, int *src_zx_gpu, float *sinus_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nsrc_gpu) return;
    p[dloc(src_zx_gpu[i],src_zx_gpu[i+nsrc_gpu])] += s_amp_gpu[i]*sinus_gpu[it-s_phs_gpu[i]+ct];
}


__global__ void inject_kernel_rec(int it, int ct, float *r_amp_gpu, int *r_phs_gpu, float *p, int *rec_zx_gpu, float *sinus_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nrec_gpu) return;
    p[dloc(rec_zx_gpu[i],rec_zx_gpu[i+nrec_gpu])] += r_amp_gpu[i]*sinus_gpu[it-r_phs_gpu[i]+ct];
}

__global__ void collect_kernel_rec(int pt, int ct, float *r_real_gpu, float *r_imag_gpu, float *p, int *rec_zx_gpu, float *sinus_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nrec_gpu) return;
    r_real_gpu[ (i)       ] += p[dloc(rec_zx_gpu[i],rec_zx_gpu[i+nrec_gpu])]*sinus_gpu[pt+ct];
    r_imag_gpu[ (i)       ] -= p[dloc(rec_zx_gpu[i],rec_zx_gpu[i+nrec_gpu])]*sinus_gpu[pt   ];
}

__global__ void collect_kernel(int pt, int ct, float *w_real_gpu, float *w_imag_gpu, float *p, float *sinus_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    if(i>=n1s_gpu || j>=n2s_gpu) return;
    w_real_gpu[i+(j)*n1s_gpu] += p[dloc(i+crop1_gpu,j+crop2_gpu)]*sinus_gpu[pt+ct];
    w_imag_gpu[i+(j)*n1s_gpu] -= p[dloc(i+crop1_gpu,j+crop2_gpu)]*sinus_gpu[pt   ];
}

/*
__global__ void collect_rec(float *r_real_gpu, float *r_imag_gpu, float *w_real_gpu, float *w_imag_gpu, int *rec_zx_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nrec_gpu) return;
    r_real_gpu[ (i) ] = w_real_gpu[dloc(rec_zx_gpu[i],rec_zx_gpu[i+nrec_gpu])];
    r_imag_gpu[ (i) ] = w_imag_gpu[dloc(rec_zx_gpu[i],rec_zx_gpu[i+nrec_gpu])];
}
*/

__global__ void collect_rec(int pt, int ct, float *w_real_gpu, float *w_imag_gpu, float *p, float *sinus_gpu)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    if(i>=nrec_gpu) return;
    w_real_gpu[i+(j)*n1s_gpu] += p[dloc(i+crop1_gpu,j+crop2_gpu)]*sinus_gpu[pt+ct];
    w_imag_gpu[i+(j)*n1s_gpu] -= p[dloc(i+crop1_gpu,j+crop2_gpu)]*sinus_gpu[pt   ];
}

extern "C" void gpu_init(int *pnt, int *nst, int *nft, float *dt, int *n1, int *n2, float *d1, float *d2, int *gpu_id, float *sinus_cpu)
{
    
     printf("Hello World");
    cudaSetDevice(*gpu_id);
    int ct  = *pnt/4;
    int snt = *pnt+*pnt+ct+1;
    int nt  = *nst+*nft;
    
    // Allocate on GPU
    cuda_call(cudaMalloc( (void**) &p0_gpu, (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &p1_gpu, (*n1)*(*n2)*sizeof(float)));
    
    //bind texture reference to array
    cf = cudaCreateChannelDesc<float>();
    cuda_call(cudaMallocArray( &vel_array, &cf, *n1, *n2));
    cuda_call(cudaMallocArray(&vel2_array, &cf, *n1, *n2));
    
    cuda_call(cudaMemcpyToSymbol(  n1gpu,  n1, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(  n2gpu,  n2, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol( nt_gpu, &nt, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(pnt_gpu, pnt, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(nst_gpu, nst, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(nft_gpu, nft, sizeof(int)));
    
    float coeffs_cpu[COEFFS_SIZE];
    coeffs_cpu[CZ1]= 3.333333333*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX1]= 3.333333333*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ2]=-.4761904762*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX2]=-.4761904762*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ3]= .0793650794*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX3]= .0793650794*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ4]=-.0099206349*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX4]=-.0099206349*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ5]= .0006349206*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX5]= .0006349206*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[C0]= coeffs_cpu[CZ5]+coeffs_cpu[CX5];
    coeffs_cpu[C0]+=coeffs_cpu[CZ4]+coeffs_cpu[CX4];
    coeffs_cpu[C0]+=coeffs_cpu[CZ3]+coeffs_cpu[CX3];
    coeffs_cpu[C0]+=coeffs_cpu[CZ2]+coeffs_cpu[CX2];
    coeffs_cpu[C0]+=coeffs_cpu[CZ1]+coeffs_cpu[CX1];
    coeffs_cpu[C0]=-2.*coeffs_cpu[C0];
    
    cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));
    
    // upload sinusoid
    cuda_call(cudaMalloc( (void**) &sinus_gpu, snt*sizeof(float)));
    cuda_call(cudaMemcpy(sinus_gpu, sinus_cpu, snt*sizeof(float), cudaMemcpyHostToDevice));

}


extern "C" void gpu_init_model(int *nrec, int *nsrc)
{
    cuda_call(cudaMemcpyToSymbol(nrec_gpu,  nrec,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(nsrc_gpu,  nsrc,   sizeof(int)));
    cuda_call(cudaMalloc( (void**) &s_phs_gpu ,   (*nsrc)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &s_amp_gpu ,   (*nsrc)*sizeof(int  )));
    cuda_call(cudaMalloc( (void**) &r_real_gpu,   (*nrec)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &r_imag_gpu,   (*nrec)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &rec_zx_gpu, 2*(*nrec)*sizeof(int  )));
    cuda_call(cudaMalloc( (void**) &src_zx_gpu, 2*(*nsrc)*sizeof(int  )));
}

extern "C" void gpu_init_prop(int *nrec, int *nsrc, int *n1s, int *n2s, int *crop1, int *crop2)
{
    gpu_init_model(nrec, nsrc);
    cuda_call(cudaMalloc( (void**) &r_phs_gpu ,       (*nrec)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &r_amp_gpu ,       (*nrec)*sizeof(int  )));

    cuda_call(cudaMalloc( (void**) &w_real_gpu, (*n1s)*(*n2s)*sizeof(float)));
    cuda_call(cudaMalloc( (void**) &w_imag_gpu, (*n1s)*(*n2s)*sizeof(float)));
    cuda_call(cudaMemcpyToSymbol(crop1_gpu,  crop1,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(crop2_gpu,  crop2,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n1s_gpu,  n1s,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2s_gpu,  n2s,   sizeof(int)));
}


extern "C" void gpu_vel_update(float *cpu_vel, float *cpu_vel2)
{
    int n1, n2;
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int))); 
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int))); 
    cuda_call(cudaMemcpyToArray( vel_array, 0,0,  cpu_vel, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel_ref, vel_array));
    cuda_call(cudaMemcpyToArray(vel2_array, 0,0, cpu_vel2, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel2_ref, vel2_array));
}


extern "C" void gpu_model(int *src_zx, int *rec_zx, float *s_amp, int *s_phs, float *r_real_cpu, float *r_imag_cpu, int *ft_j)
{
    int n1, n2, nrec, nsrc, pnt, nft, nst, ct, pt, nt;
    float *ptemp;

    // Download some dimensions from the initialized gpu
    cuda_call(cudaMemcpyFromSymbol(&pnt, pnt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nst, nst_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nft, nft_gpu, sizeof(int)));
    nt  = nst+nft;
    ct  = pnt/4;

    cuda_call(cudaMemcpyFromSymbol(  &n1, n1gpu   , sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(  &n2, n2gpu   , sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nsrc, nsrc_gpu, sizeof(int)));
    
    // Zero wavefields, source functions and data on gpu 
    cuda_call(cudaMemset(    p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(    p1_gpu, 0,n1*n2*sizeof(float)));
    
    cuda_call(cudaMemset( s_phs_gpu, 0, nsrc*sizeof(int  )));
    cuda_call(cudaMemset( s_amp_gpu, 0, nsrc*sizeof(float)));
    cuda_call(cudaMemset(r_real_gpu, 0, nrec*sizeof(float)));
    cuda_call(cudaMemset(r_imag_gpu, 0, nrec*sizeof(float)));
    
    // Upload source and receiver positions
    cuda_call(cudaMemcpy(src_zx_gpu, src_zx,2*nsrc*sizeof(int), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(rec_zx_gpu, rec_zx,2*nrec*sizeof(int), cudaMemcpyHostToDevice));
    
    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
        
    // Upload source amplitudes and phases
    cuda_call(cudaMemcpy(s_phs_gpu, s_phs, nsrc*sizeof(int  ), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(s_amp_gpu, s_amp, nsrc*sizeof(float), cudaMemcpyHostToDevice));
    
    
    /* 11 ACTUAL WAVE PROPAGATION */
    for (int it=0; it<nt; it++)
    {
        // One step
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;  // Iterate wavefields

        // Find (t=0)+it index in uploaded sinusoid      
        pt = it % pnt + pnt/2;
        
        // Inject source
        kernel_exec  ( inject_kernel_src<<<nsrc/80+1,80>>>( pt, ct,  s_amp_gpu,  s_phs_gpu, p1_gpu, src_zx_gpu, sinus_gpu ));
        if(it>nst & it%(*ft_j)==0)
          {
          // IFT data-lag
          kernel_exec(collect_kernel_rec<<<nrec/80+1,80>>>( pt, ct, r_real_gpu, r_imag_gpu, p1_gpu, rec_zx_gpu, sinus_gpu ));
          }
    }
    
    // Download data
    cuda_call(cudaMemcpy( r_real_cpu, r_real_gpu, nrec*sizeof(float), cudaMemcpyDeviceToHost));
    cuda_call(cudaMemcpy( r_imag_cpu, r_imag_gpu, nrec*sizeof(float), cudaMemcpyDeviceToHost));

}


extern "C" void gpu_prop_for(int *src_zx, float *s_amp, int *s_phs, float *w_real_cpu, float *w_imag_cpu, int *ft_j)
{

    int n1, n2, nsrc, pnt, nft, nst, ct, pt, nt;
    int n1s, n2s;
    float *ptemp;
    
    // Download some dimensions from the initialized gpu
    cuda_call(cudaMemcpyFromSymbol(&pnt, pnt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nst, nst_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nft, nft_gpu, sizeof(int)));
    nt  = nst+nft;
    ct  = pnt/4;

    cuda_call(cudaMemcpyFromSymbol(&n1 ,     n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2 ,     n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1s,   n1s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2s,   n2s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nsrc, nsrc_gpu, sizeof(int)));

    // Zero wavefields, source functions and data on gpu 
    cuda_call(cudaMemset(    p0_gpu, 0,n1 *n2 *sizeof(float)));
    cuda_call(cudaMemset(    p1_gpu, 0,n1 *n2 *sizeof(float)));
    cuda_call(cudaMemset(w_real_gpu, 0,n1s*n2s*sizeof(float)));
    cuda_call(cudaMemset(w_imag_gpu, 0,n1s*n2s*sizeof(float)));
    
    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
    
    int nblocksx2=n1s/BLOCK_SIZE+1;
    int nblocksy2=n2s/BLOCK_SIZE+1;
    dim3 dimGrid2(nblocksx2,nblocksy2);
    dim3 dimBlock2(BLOCK_SIZE, BLOCK_SIZE);
    
    // Upload source positions
    cuda_call(cudaMemcpy(src_zx_gpu, src_zx, 2*nsrc*sizeof(int), cudaMemcpyHostToDevice));

    // Upload source amplitudes and phases
    cuda_call(cudaMemcpy(s_phs_gpu, s_phs, nsrc*sizeof(int  ), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(s_amp_gpu, s_amp, nsrc*sizeof(float), cudaMemcpyHostToDevice));
    
    /* 11 ACTUAL WAVE PROPAGATION */
    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;

        // Find (t=0)+it index in uploaded sinusoid      
        pt = it % pnt + pnt/2;

        // Inject source
        kernel_exec  ( inject_kernel_src<<<nsrc/80+1,80>>>( pt, ct, s_amp_gpu , s_phs_gpu , p1_gpu, src_zx_gpu, sinus_gpu ));      

        // IFT data-lag
        if(it>nst & it%(*ft_j)==0)
        {
            kernel_exec(collect_kernel<<<dimGrid2, dimBlock2>>>(pt, ct, w_real_gpu, w_imag_gpu, p1_gpu, sinus_gpu));
        }
    
    }

    // Download wavefield
    cuda_call(cudaMemcpy( w_real_cpu, w_real_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
    cuda_call(cudaMemcpy( w_imag_cpu, w_imag_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));

}



extern "C" void gpu_prop_adj(int *rec_zx, float *r_amp, int *r_phs, float *w_real_cpu, float *w_imag_cpu, int *ft_j)
{

    int n1, n2, nrec, pnt, nft, nst, ct, pt, nt;
    int n1s, n2s;
    float *ptemp;
    
    // Download some dimensions from the initialized gpu
    cuda_call(cudaMemcpyFromSymbol(&pnt, pnt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nst, nst_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nft, nft_gpu, sizeof(int)));
    nt  = nst+nft;
    ct  = pnt/4;

    cuda_call(cudaMemcpyFromSymbol(&n1 ,     n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2 ,     n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1s,   n1s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2s,   n2s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));

    // Zero wavefields, source functions and data on gpu 
    cuda_call(cudaMemset(    p0_gpu, 0,n1 *n2 *sizeof(float)));
    cuda_call(cudaMemset(    p1_gpu, 0,n1 *n2 *sizeof(float)));
    cuda_call(cudaMemset(w_real_gpu, 0,n1s*n2s*sizeof(float)));
    cuda_call(cudaMemset(w_imag_gpu, 0,n1s*n2s*sizeof(float)));
    
    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
    
    int nblocksx2=n1s/BLOCK_SIZE+1;
    int nblocksy2=n2s/BLOCK_SIZE+1;
    dim3 dimGrid2(nblocksx2,nblocksy2);
    dim3 dimBlock2(BLOCK_SIZE, BLOCK_SIZE);
    
    // Upload source positions
    cuda_call(cudaMemcpy(rec_zx_gpu,rec_zx,2*nrec*sizeof(int), cudaMemcpyHostToDevice));

    // Upload source amplitudes and phases
    cuda_call(cudaMemcpy(r_phs_gpu, r_phs, nrec*sizeof(int  ), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(r_amp_gpu, r_amp, nrec*sizeof(float), cudaMemcpyHostToDevice));
    
    /* 11 ACTUAL WAVE PROPAGATION */
    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;

        // Find (t=0)+it index in uploaded sinusoid      
        pt = it % pnt + pnt/2;

        // Inject source
        kernel_exec  ( inject_kernel_rec<<<nrec/80+1,80>>>( pt, ct, r_amp_gpu , r_phs_gpu , p1_gpu, rec_zx_gpu, sinus_gpu ));

        // IFT data-lag
        if(it>nst & it%(*ft_j)==0)
        {
            kernel_exec(collect_kernel<<<dimGrid2, dimBlock2>>>(pt, ct, w_real_gpu, w_imag_gpu, p1_gpu, sinus_gpu));
        }
    
    }

    // Download wavefield
    cuda_call(cudaMemcpy( w_real_cpu, w_real_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
    cuda_call(cudaMemcpy( w_imag_cpu, w_imag_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));

}



//    for (int ir=0; ir<2*(*nrec); ir++)
//    {printf("%d ",rec_zx[ir]);}

//    printf( "nbx %d  nby %d \n", nblocksx,nblocksy);
