module tomo_mod

implicit none

contains

!!---------------------------------------------------------- 

subroutine SecDer1d(souIn, souOut)
  !! take 2nd derivative
  integer                       :: i
  real, dimension(:)            :: souIn, souOut
  real, dimension(size(souIn,1))  :: temp

  temp = souIn
  do i=2,size(souIn,1)-1
    souOut(i) = 2*temp(i) - temp(i+1) - temp(i-1)
  end do

  souOut(1) = 0.
  souOut(size(souIn,1)) = 0.

end subroutine

!!---------------------------------------------------------- 

subroutine extend(gradient, vel_nz, vel_nx, crop1, crop2)
  !! compute absorbing boundary condition weights
  integer                       :: crop1, crop2
  real, dimension(:,:)          :: gradient
  integer                       :: iz, ix, vel_nz, vel_nx

  do ix=1,crop2
    gradient(:,ix) = gradient(:,1+crop2)
  enddo
  do ix=vel_nx-crop2+1,vel_nx
    gradient(:,ix) = gradient(:,vel_nx-crop2)
  enddo
  do iz=1,crop1
    gradient(iz,:) = gradient(1+crop1,:)
  enddo
  do iz=vel_nz-crop1+1,vel_nz
    gradient(iz,:) = gradient(vel_nz-crop1,:)
  enddo

end subroutine

!!---------------------------------------------------------- 

subroutine stepper(r0, r1, r2, r3, step_size, ismall, stat)
  !! testing step sizes
  double precision              :: r0, r1, r2, r3, step_size
  integer                       :: stat, ismall

  if(isnan(r1) .or. r1 .eq. r1 + r1) then
    write(0,*) "something wrong with r1, will reset its value"
    r1 = r0 + abs(r0)
  endif
  if(isnan(r2) .or. r2 .eq. r2 + r2) then
    write(0,*) "something wrong with r2, will reset its value"
    r2 = r0 + abs(r0)
  endif
  if(isnan(r3) .or. r3 .eq. r3 + r3) then
    write(0,*) "something wrong with r3, will reset its value"
    r3 = r0 + abs(r0)
  endif

  if(r0 < r1 .and. r0 < r2 .and. r0 < r3) then
    write(0,*) "all r's are larger than r0, will try dividing by 4"
    step_size = step_size * 0.25
    stat = 1
  elseif(r1 < r2 .and. r1 < r3) then
    write(0,*) "step scalar is = 0.5"
    step_size = step_size * 0.5
    write(0,*) "The best step size is = ", step_size
    ismall = 1
    stat = 0
  elseif(r2 < r1 .and. r2 < r3) then
    write(0,*) "step scalar is = 1.0"
    write(0,*) "The best step size is = ", step_size
    ismall = 2
    stat = 0
  elseif(r3 < r1 .and. r3 < r2) then
    write(0,*) "step scalar is = 1.5"
    step_size = step_size * 1.5
    write(0,*) "The best step size is = ", step_size
    ismall = 3
    stat = 0
  else
    write(0,*) "unknown step size problem, will try dividing by 4"
    step_size = step_size * 0.25
    stat = 1
  endif

end subroutine


end module
