/* 2D Time domain finite difference kernel
	By: Abdullah AlTheyab
	Ref: similar to Paulius code but for 2D
 */
#define BLOCK_SIZE 16
#define FAT 5
#define COEFFS_SIZE 11

#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4
#define CZ3 5
#define CX3 6
#define CZ4 7
#define CX4 8
#define CZ5 9
#define CX5 10

__constant__ float coeffs[COEFFS_SIZE];

extern "C" __global__ void wave_kernel(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
	int ig=FAT+threadIdx.x+blockIdx.x*BLOCK_SIZE; 	// global z coordinates
	int jg=FAT+threadIdx.y+blockIdx.y*BLOCK_SIZE; 	// global x coordinates
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	int addr=dloc(ig,jg);		  		  			// linear memroy address

	//copy to shared memory
	p1s[jl][il]=p1[addr];

	if(threadIdx.y<FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
	}
	if(threadIdx.x<FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
	}
	
	__syncthreads();

	p2[addr]= tex2D(vel2_ref, ig,jg)*(
			coeffs[CX5]*(p1s[jl-5][il]+p1s[jl+5][il])
			+coeffs[CZ5]*(p1s[jl][il-5]+p1s[jl][il+5])
			+coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
			+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
			+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
			+coeffs[C0]*p1s[jl][il]
			)+p1s[jl][il]+p1s[jl][il]-p0[addr];

}
